// Copyright 2020 Canonical Ltd.
// Licensed under the LGPLv3 with static-linking exception.
// See LICENCE file for details.

package efi

type PhysicalAddress uint64

// LBA corresponds to the EFI_LBA type.
type LBA uint64
