// -*- Mode: Go; indent-tabs-mode: t -*-
//go:build !linux

/*
 * Copyright (C) 2017-2026 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package systemd

import (
	"errors"
	"os"
)

var errUnsupported = errors.New("unsupported on non-Linux systems")

var osGetenv = os.Getenv

func SdNotify(notifyState string) error {
	return errUnsupported
}

func SdNotifyWithFds(notifyState string, files ...*os.File) error {
	return errUnsupported
}
