// Copyright 2021 Canonical Ltd.
// Licensed under the LGPLv3 with static-linking exception.
// See LICENCE file for details.

/*
Package cryptutil contains some cryptographic functions that are useful when using go-tpm2.
*/
package cryptutil
