// -*- Mode: Go; indent-tabs-mode: t -*-

/*
 * Copyright (C) 2017 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package backend

import (
	"github.com/snapcore/snapd/interfaces/mount"
)

// Discard the mount namespace of a given snap.
func (b Backend) DiscardSnapNamespace(snapName string) error {
	return mount.DiscardSnapNamespace(snapName)
}

// Discard the mount namespace of a locked snap, e.g. during a refresh when
// running applications is temporarily inhibited.
func (b Backend) DiscardLockedSnapNamespace(snapName string) error {
	return mount.DiscardLockedSnapNamespace(snapName)
}
