// Copyright 2019 Canonical Ltd.
// Licensed under the LGPLv3 with static-linking exception.
// See LICENCE file for details.

/*
Package linux provides an interface for communicating with TPMs using a Linux TPM character device
*/
package linux
