// Copyright 2023 Canonical Ltd.
// Licensed under the LGPLv3 with static-linking exception.
// See LICENCE file for details.

/*
Package policyutil contains utilties for constructing and executing authorization policies.
*/
package policyutil
