// Copyright 2023 Canonical Ltd.
// Licensed under the LGPLv3 with static-linking exception.
// See LICENCE file for details.

/*
Package objectutil contains utilities for creating and working with objects.
*/
package objectutil
