//go:build go1.20
// +build go1.20

// Copyright 2025 Canonical Ltd.
// Licensed under the LGPLv3 with static-linking exception.
// See LICENCE file for details.

package tpm2

import "fmt"

var formatString = fmt.FormatString
